from http.cookies import SimpleCookie


class CookieUtil:
    @classmethod
    def build_set_cookie_header(
        cls,
        name: str,
        domain: str,
        value: str,
        path: str,
        expires_in_seconds: int,
        is_secure: bool = True,
        is_httponly: bool = False,
    ) -> str:
        cookie = SimpleCookie()
        cookie[name] = value
        cookie[name]['domain'] = domain
        cookie[name]['path'] = path
        cookie[name]['expires'] = expires_in_seconds
        cookie[name]['secure'] = is_secure
        cookie[name]['httponly'] = is_httponly

        return cookie.output(header='', sep='\015\012').strip()



from py_aws_core import decorators, dynamodb_entities, exceptions, logs
from py_aws_core.boto_responses import ItemResponse, UpdateItemResponse
from py_aws_core.dynamodb_api import DynamoDBAPI

logger = logs.get_logger()


class GetOrCreateSession(DynamoDBAPI):
    class Response(UpdateItemResponse):
        @property
        def session(self) -> dynamodb_entities.Session:
            return dynamodb_entities.Session(data=self.attributes)

    @classmethod
    @decorators.dynamodb_handler(client_err_map=exceptions.ERR_CODE_MAP, cancellation_err_maps=[])
    def call(
        cls,
        table,
        session_id: str,
        created_at_datetime: str,
        expires_at: int = None
    ):
        pk = sk = dynamodb_entities.Session.create_key(_id=session_id)
        _type = dynamodb_entities.Session.type()
        update_fields = [
            cls.UpdateField(expression_attr='ty'),
            cls.UpdateField(expression_attr='si'),
            cls.UpdateField(expression_attr='ma'),
            cls.UpdateField(expression_attr='ea', set_once=True),
            cls.UpdateField(expression_attr='ca', set_once=True),
        ]
        response = table.update_item(
            Key={
                'PK': pk,
                'SK': sk,
            },
            UpdateExpression=cls.build_update_expression(update_fields),
            ExpressionAttributeNames={
                '#ty': 'Type',
                "#si": 'SessionId',
                '#ca': 'CreatedAt',
                '#ma': 'ModifiedAt',
                '#ea': 'ExpiresAt',
            },
            ExpressionAttributeValues={
                ':ty': _type,
                ':si': session_id,
                ':ca': created_at_datetime,
                ':ma': created_at_datetime,
                ':ea': expires_at,
            },
            ReturnValues='ALL_NEW'
        )

        logger.debug(f'GetOrCreateSession called', response=response)
        return cls.Response(response)


class GetSessionItem(DynamoDBAPI):
    class Response(ItemResponse):
        @property
        def session(self) -> dynamodb_entities.Session:
            return dynamodb_entities.Session(data=self.Item)

    @classmethod
    @decorators.dynamodb_handler(client_err_map=exceptions.ERR_CODE_MAP, cancellation_err_maps=[])
    def call(
        cls,
        table,
        session_id: str
    ) -> Response:
        pk = sk = dynamodb_entities.Session.create_key(_id=session_id)
        response = table.get_item(
            Key={
                'PK': pk,
                'SK': sk
            },
            ExpressionAttributeNames={
                "#pk": "PK",
                "#bc": "Base64Cookies",
                "#tp": "Type"
            },
            ProjectionExpression='#pk, #bc, #tp'
        )
        logger.debug(f'GetSessionItem called', response=response)
        return cls.Response(response)


class PutSession(DynamoDBAPI):
    @classmethod
    @decorators.dynamodb_handler(client_err_map=exceptions.ERR_CODE_MAP, cancellation_err_maps=[])
    def call(
        cls,
        table,
        session_id: str,
        b64_cookies: bytes
    ):
        pk = sk = dynamodb_entities.Session.create_key(_id=session_id)
        _type = dynamodb_entities.Session.type()
        item = cls.get_put_item_map(
            pk=pk,
            sk=sk,
            _type=_type,
            expire_in_seconds=None,
            Base64Cookies=b64_cookies,
            SessionId=session_id
        )
        response = table.put_item(
            Item=item,
        )
        logger.debug(f'PutSession called', response=response)
        return response


class UpdateSessionCookies(DynamoDBAPI):
    class Response(UpdateItemResponse):
        @property
        def session(self) -> dynamodb_entities.Session:
            return dynamodb_entities.Session(self.attributes)

    @classmethod
    @decorators.dynamodb_handler(client_err_map=exceptions.ERR_CODE_MAP, cancellation_err_maps=[])
    def call(
        cls,
        table,
        session_id: str,
        b64_cookies: bytes,
        now_datetime: str
    ):
        pk = sk = dynamodb_entities.Session.create_key(_id=session_id)
        response = table.update_item(
            Key={
                'PK': pk,
                'SK': sk,
            },
            UpdateExpression='SET #b64 = :b64, #mda = :mda',
            ExpressionAttributeNames={
                '#b64': 'Base64Cookies',
                '#mda': 'ModifiedAt',
            },
            ExpressionAttributeValues={
                ':b64': b64_cookies,
                ':mda': now_datetime
            },
            ReturnValues='ALL_NEW'
        )
        logger.debug(f'UpdateSessionCookies called', response=response)
        return cls.Response(response)



from py_aws_core.encoders import JsonEncoder


class AsDictMixin:
    """
    Converts
    """
    def as_dict(self):
        return {k: v for k, v in self.__dict__.items() if not k.startswith('_')}


class JsonMixin:
    @property
    def to_json(self):
        return JsonEncoder().serialize_to_json(self)



from unittest import mock

from py_aws_core.spoofing.proxyrack import const
from py_aws_core.spoofing.proxyrack.proxyrack_service import ProxyRackService
from py_aws_core.testing import BaseTestFixture


class ProxyRackProxyBackendTests(BaseTestFixture):
    """
        ProxyRackProxyBackend Tests
    """

    def test_proxy_url(self):
        proxy_service = ProxyRackService(proxy_username='user123', proxy_password='pass456')
        proxy_url = proxy_service.get_proxy_url(
            cities=['Dallas'],
            netloc='megaproxy.rotating.proxyrack.net:10000',
            proxy_ip='192.168.86.250',
            proxy_os=const.ProxyOs.WINDOWS,
            session_id='user123',
        )

        self.assertEqual(
            'http://user123;city=Dallas;osName=Windows;proxyIp=192.168.86.250;session=user123;refreshMinutes=60:pass456@megaproxy.rotating.proxyrack.net:10000',
            proxy_url
        )



import json

from botocore.client import BaseClient
from botocore.exceptions import ClientError

from . import exceptions, logs, utils
from .secrets_interface import ISecrets

logger = logs.get_logger()


class SSMParameterStore(ISecrets):
    """
        First checks environment variables for secrets.
        Second, checks cached secrets
        Third, if secret not found, will attempt to pull from secrets manager
    """
    class Response:
        class Parameter:
            def __init__(self, data):
                self.Name = data['Name']
                self.Type = data['Type']
                self.Value = data['Value']
                self.Version = data['Version']
                self.Selector = data['Selector']
                self.SourceResult = data['SourceResult']
                self.LastModifiedDate = data['LastModifiedDate']
                self.ARN = data['ARN']
                self.DataType = data['DataType']

        def __init__(self, data):
            self.Parameter = self.Parameter(data['Parameter'])

        @property
        def parameter_json(self):
            return json.loads(self.Parameter.Value)

    AWS_SECRET_ID_KEY = 'AWS_SECRET_NAME'

    def __init__(self, boto_client: BaseClient, cached_secrets: dict = None):
        self._boto_client = boto_client
        self._cached_secrets = cached_secrets or dict()

    def get_secret(self, secret_name: str) -> str:
        if secret_value := utils.get_environment_variable(secret_name):
            logger.debug(f'Secret "{secret_name}" found in environment variables')
            return secret_value
        if val := self._cached_secrets.get(secret_name):
            logger.debug(f'Secret "{secret_name}" found in cached secrets')
            return val
        try:
            r_get_parameter = self._boto_client.get_parameter(Name=self.aws_secret_id)
            r_get_parameter = self.Response(r_get_parameter)
            self._cached_secrets = r_get_parameter.parameter_json
            return self._cached_secrets[secret_name]
        except ClientError as e:
            logger.exception(f'Error while trying to find secret "{secret_name}"')
            # For a list of exceptions thrown, see
            # https://docs.aws.amazon.com/secretsmanager/latest/apireference/API_GetSecretValue.html
            raise exceptions.SecretsManagerException(e)

    @property
    def aws_secret_id(self) -> str:
        if aws_secret_id := utils.get_environment_variable(self.AWS_SECRET_ID_KEY):
            return aws_secret_id
        raise exceptions.SecretsManagerException(f'Missing environment variable "{self.AWS_SECRET_ID_KEY}"')



import json
from py_aws_core import logs
from decimal import Decimal

logger = logs.get_logger()


class JsonEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, set):
            return list(obj)
        elif isinstance(obj, Decimal):
            return float(obj)
        elif attrs := getattr(obj, '__dict__', None):  # Omit any non-public attributes
            return {k: v for k, v in attrs.items() if '__' not in k}
        else:
            super().default(obj)

    @classmethod
    def serialize_to_json(cls, obj):
        return json.dumps(obj, cls=cls, separators=(',', ':'), ensure_ascii=False)



from dataclasses import dataclass

import boto3
from boto3.dynamodb.table import TableResource
from boto3.dynamodb.types import TypeDeserializer, TypeSerializer

from py_aws_core import const, logs, utils

logger = logs.get_logger()


class DynamoDBAPI:
    @dataclass
    class UpdateField:
        expression_attr: str
        set_once: bool = False

    @classmethod
    def build_update_expression(cls, fields: list[UpdateField]) -> str:
        n_fields = [f'#{f.expression_attr} = :{f.expression_attr}' for f in fields if not f.set_once]
        o_fields = [f'#{f.expression_attr} = if_not_exists(#{f.expression_attr}, :{f.expression_attr})' for f in fields
                    if f.set_once]
        return f'SET {', '.join(n_fields + o_fields)}'

    @classmethod
    def get_batch_entity_create_map(
        cls,
        pk: str,
        sk: str,
        _type: str,
        created_by: str = '',
        expire_in_seconds: int | None = const.DB_DEFAULT_EXPIRES_IN_SECONDS,
        **kwargs,
    ) -> dict:
        return {
            'PK': pk,
            'SK': sk,
            'Type': _type,
            'CreatedAt': utils.to_iso_8601(),
            'CreatedBy': created_by,
            'ModifiedAt': '',
            'ModifiedBy': '',
            'ExpiresAt': cls.calc_expire_at_timestamp(expire_in_seconds=expire_in_seconds),
        } | kwargs

    @classmethod
    def get_entity_update_map(
        cls,
        pk: str,
        sk: str,
        modified_by: str = '',
        **kwargs,
    ) -> dict:
        return {
            'PK': pk,
            'SK': sk,
            'ModifiedAt': utils.to_iso_8601(),
            'ModifiedBy': modified_by,
        } | kwargs

    @classmethod
    def get_put_item_map(
        cls,
        pk: str,
        sk: str,
        _type: str,
        created_by: str = '',
        expire_in_seconds: int | None = const.DB_DEFAULT_EXPIRES_IN_SECONDS,
        **kwargs,
    ) -> dict:
        return {
           'PK': pk,
           'SK': sk,
           'Type': _type,
           'CreatedAt': utils.to_iso_8601(),
           'CreatedBy': created_by,
           'ModifiedAt': '',
           'ModifiedBy': '',
           'ExpiresAt': cls.calc_expire_at_timestamp(expire_in_seconds=expire_in_seconds),
        } | kwargs

    @classmethod
    def batch_write_item_maps(cls, table_resource: TableResource, item_maps: list[dict]) -> int:
        with table_resource.batch_writer() as batch:
            for _map in item_maps:
                batch.put_item(Item=_map)
        return len(item_maps)

    @classmethod
    def get_new_table_resource(cls, table_name: str):
        dynamodb_resource = boto3.resource('dynamodb')
        return dynamodb_resource.Table(table_name)

    @classmethod
    def serialize_types(cls, data: dict):
        """
        Converts normalized json to low level dynamo json
        """
        return {k: TypeSerializer().serialize(v) for k, v in data.items()}

    @classmethod
    def deserialize_types(cls, data: dict):
        """
        Converts low level dynamo json to normalized json
        """
        return {k: TypeDeserializer().deserialize(v) for k, v in data.items()}

    @classmethod
    def calc_expire_at_timestamp(cls, expire_in_seconds: int = None) -> int | str:
        """
        Adds seconds to current unix timestamp to generate new unix timestamp
        Seconds set to None will result in empty string
        :param expire_in_seconds: Seconds to add to current timestamp
        :return:
        """
        if expire_in_seconds is None:
            return ''
        return utils.add_seconds_to_current_unix_timestamp(seconds=expire_in_seconds)



from abc import abstractmethod


class ISecrets:
    @abstractmethod
    def get_secret(self, secret_name: str) -> str:
        pass


class IDynamoDBSecrets(ISecrets):
    @abstractmethod
    def get_table_name(self) -> str:
        pass



from py_aws_core.testing import BaseTestFixture
from py_aws_core.spoofing.proxyrack import const, utils


class UtilTests(BaseTestFixture):
    """
    Utility Tests
    """

    def test_proxy_builder(self):
        config = utils.ProxyBuilder.Config(
            cities=['Seattle', 'New York', 'Los Angeles'],
            country='US',
            isps=['Verizon', 'Ipxo Limited'],
            proxy_ip='184.53.48.165',
            proxy_os=const.ProxyOs.LINUX,
            refresh_minutes=10,
            session_id='13ac97fe-0f26-45ff-aeb9-2801400326ec',
            missingKey=''
        )
        proxy_builder = utils.ProxyBuilder(
            username='proxyman123',
            password='goodpw567',
            netloc='megaproxy.rotating.proxyrack.net:10000',
            config=config
        )

        self.assertEqual(
            proxy_builder.http_url,
            'http://proxyman123;city=Seattle,NewYork,LosAngeles;country=US;isp=Verizon,IpxoLimited;osName=Linux;proxyIp=184.53.48.165;session=13ac97fe-0f26-45ff-aeb9-2801400326ec;refreshMinutes=10:goodpw567@megaproxy.rotating.proxyrack.net:10000'
        )



from abc import ABC, abstractmethod

from . import dynamodb_entities


class IDatabase(ABC):
    @classmethod
    @abstractmethod
    def get_or_create_session(cls, session_id: str) -> dynamodb_entities.Session:
        pass

    @classmethod
    @abstractmethod
    def get_session_item(cls, session_id: str) -> dynamodb_entities.Session:
        pass

    @classmethod
    @abstractmethod
    def put_session_item(cls, session_id: str, b64_cookies: bytes):
        pass

    @classmethod
    @abstractmethod
    def update_session_cookies(cls, session_id: str, b64_cookies: bytes) -> dynamodb_entities.Session:
        pass



import json
from importlib.resources import as_file

import respx

from py_aws_core.clients import RetryClient
from py_aws_core.testing import BaseTestFixture
from py_aws_core.spoofing.proxyrack import exceptions, proxyrack_api
from . import const as test_const


class ActiveConnectionsTests(BaseTestFixture):
    """
        Active Connections Tests
    """

    @respx.mock
    def test_ok(self):
        source = test_const.TEST_RESOURCE_PATH.joinpath('active_conns.json')
        with as_file(source) as active_conns_json:
            mocked_route_active_conns = self.create_ok_route(
                method='GET',
                url__eq='http://api.proxyrack.net/active_conns',
                _json=json.loads(active_conns_json.read_text(encoding='utf-8'))
            )

        with RetryClient() as client:
            r_active_connections = proxyrack_api.GetActiveConnections.call(client)
            self.assertEqual(2, len(r_active_connections.connections))

        self.assertEqual(1, mocked_route_active_conns.call_count)

    @respx.mock
    def test_ProxyRackError(self):
        source = test_const.TEST_RESOURCE_PATH.joinpath('active_conns.json')
        with as_file(source) as active_conns_json:
            mocked_route_active_conns = self.create_route(
                method='GET',
                url__eq='http://api.proxyrack.net/active_conns',
                response_status_code=400,
                response_json=json.loads(active_conns_json.read_text(encoding='utf-8'))
            )

        with RetryClient() as client:
            with self.assertRaises(exceptions.ProxyRackException):
                proxyrack_api.GetActiveConnections.call(client)

        self.assertEqual(1, mocked_route_active_conns.call_count)


class PostTempAPIKeyTests(BaseTestFixture):
    """
        API Keys Tests
    """

    @respx.mock
    def test_ok(self):
        source = test_const.TEST_RESOURCE_PATH.joinpath('passwords.json')
        with as_file(source) as passwords_json:
            mocked_route_passwords = self.create_route(
                method='POST',
                url__eq='http://api.proxyrack.net/passwords?expirationSeconds=60',
                response_status_code=200,
                response_json=json.loads(passwords_json.read_text(encoding='utf-8'))
            )

        with RetryClient() as client:
            r_generate_temp_api_key = proxyrack_api.PostTempAPIKey.call(client, expiration_seconds=60)
            self.assertEqual('temp-bf3702-be83a4-0bbfc1-be7f58-62cfff', r_generate_temp_api_key.api_key)

        self.assertEqual(1, mocked_route_passwords.call_count)

    @respx.mock
    def test_ProxyRackError(self):
        source = test_const.TEST_RESOURCE_PATH.joinpath('passwords.json')
        with as_file(source) as passwords_json:
            mocked_route_passwords = self.create_route(
                method='POST',
                url__eq='http://api.proxyrack.net/passwords?expirationSeconds=60',
                response_status_code=400,
                response_json=json.loads(passwords_json.read_text(encoding='utf-8'))
            )

        with RetryClient() as client:
            with self.assertRaises(exceptions.ProxyRackException):
                proxyrack_api.PostTempAPIKey.call(
                    client,
                    expiration_seconds=60
                )
        self.assertEqual(1, mocked_route_passwords.call_count)


class StatsTests(BaseTestFixture):
    """
        Stats Tests
    """

    @respx.mock
    def test_ok(self):
        source = test_const.TEST_RESOURCE_PATH.joinpath('stats.json')
        with as_file(source) as stats_json:
            mocked_route_stats = self.create_route(
                method='GET',
                url__eq='http://api.proxyrack.net/stats',
                response_status_code=200,
                response_json=json.loads(stats_json.read_text(encoding='utf-8'))
            )

        with RetryClient() as client:
            r_stats = proxyrack_api.GetStats.call(client=client)
            self.assertEqual(r_stats.thread_limit, 10000)

        self.assertEqual(mocked_route_stats.call_count, 1)

    @respx.mock
    def test_400(self):
        source = test_const.TEST_RESOURCE_PATH.joinpath('stats.json')
        with as_file(source) as stats_json:
            mocked_route_stats = self.create_route(
                method='GET',
                url__eq='http://api.proxyrack.net/stats',
                response_status_code=400,
                response_json=json.loads(stats_json.read_text(encoding='utf-8'))
            )

        with self.assertRaises(exceptions.ProxyRackException):
            with RetryClient() as client:
                proxyrack_api.GetStats.call(client=client)

        self.assertTrue(mocked_route_stats.called)


class IspsTests(BaseTestFixture):
    """
        Isps Tests
    """

    @respx.mock
    def test_ok(self):
        source = test_const.TEST_RESOURCE_PATH.joinpath('us_isps.json')
        with as_file(source) as us_isps_json:
            mocked_route_us_isps = self.create_route(
                method='GET',
                url__eq='http://api.proxyrack.net/countries/US/isps',
                response_status_code=200,
                response_json=json.loads(us_isps_json.read_text(encoding='utf-8'))
            )

        with RetryClient() as client:
            r_isps = proxyrack_api.GetISPs.call(client=client, country='US')
            self.assertEqual(len(r_isps.isps), 54)

        self.assertEqual(mocked_route_us_isps.call_count, 1)

    @respx.mock
    def test_400(self):
        source = test_const.TEST_RESOURCE_PATH.joinpath('us_isps.json')
        with as_file(source) as us_isps_json:
            mocked_route_us_isps = self.create_route(
                method='GET',
                url__eq='http://api.proxyrack.net/countries/US/isps',
                response_status_code=400,
                response_json=json.loads(us_isps_json.read_text(encoding='utf-8'))
            )

        with self.assertRaises(exceptions.ProxyRackException):
            with RetryClient() as client:
                proxyrack_api.GetISPs.call(client=client, country='US')

        self.assertEqual(mocked_route_us_isps.call_count, 1)


class CountriesTests(BaseTestFixture):
    """
        Countries Tests
    """

    @respx.mock
    def test_ok(self):
        source = test_const.TEST_RESOURCE_PATH.joinpath('countries.json')
        with as_file(source) as countries_json:
            mocked_route_countries = self.create_route(
                method='GET',
                url__eq='http://api.proxyrack.net/countries',
                response_status_code=200,
                response_json=json.loads(countries_json.read_text(encoding='utf-8'))
            )

        with RetryClient() as client:
            r_countries = proxyrack_api.GetCountries.call(client=client)
            self.assertEqual(len(r_countries.countries), 205)

        self.assertEqual(mocked_route_countries.call_count, 1)

    @respx.mock
    def test_400(self):
        source = test_const.TEST_RESOURCE_PATH.joinpath('countries.json')
        with as_file(source) as countries_json:
            mocked_route_countries = self.create_route(
                method='GET',
                url__eq='http://api.proxyrack.net/countries',
                response_status_code=400,
                response_json=json.loads(countries_json.read_text(encoding='utf-8'))
            )

        with self.assertRaises(exceptions.ProxyRackException):
            with RetryClient() as client:
                proxyrack_api.GetCountries.call(client=client)

        self.assertEqual(mocked_route_countries.call_count, 1)


class CitiesTests(BaseTestFixture):
    """
        Cities Tests
    """

    @respx.mock
    def test_ok(self):
        source = test_const.TEST_RESOURCE_PATH.joinpath('cities.json')
        with as_file(source) as cities_json:
            mocked_route_cities = self.create_route(
                method='GET',
                url__eq='http://api.proxyrack.net/cities',
                response_status_code=200,
                response_json=json.loads(cities_json.read_text(encoding='utf-8'))
            )

        with RetryClient() as client:
            r_cities = proxyrack_api.GetCities.call(client=client)

        self.assertEqual(len(r_cities.cities), 82)
        self.assertEqual(mocked_route_cities.call_count, 1)

    @respx.mock
    def test_400(self):
        source = test_const.TEST_RESOURCE_PATH.joinpath('cities.json')
        with as_file(source) as cities_json:
            mocked_route_cities = self.create_route(
                method='GET',
                url__eq='http://api.proxyrack.net/cities',
                response_status_code=400,
                response_json=json.loads(cities_json.read_text(encoding='utf-8'))
            )

        with self.assertRaises(exceptions.ProxyRackException):
            with RetryClient() as client:
                proxyrack_api.GetCities.call(client=client)

        self.assertEqual(mocked_route_cities.call_count, 1)


class CountryIPCountTests(BaseTestFixture):
    """
        Cities Tests
    """

    @respx.mock
    def test_ok(self):
        mocked_route_ip_count = self.create_route(
            method='GET',
            url__eq='http://api.proxyrack.net/countries/US/count',
            response_status_code=200,
            response_text='152'
        )

        with RetryClient() as client:
            r_country_ip_count = proxyrack_api.GetCountryIPCount.call(client=client, country='US')
            self.assertEqual(r_country_ip_count, '152')

        self.assertEqual(mocked_route_ip_count.call_count, 1)

    @respx.mock
    def test_400(self):
        mocked_route_ip_count = self.create_route(
            method='GET',
            url__eq='http://api.proxyrack.net/countries/US/count',
            response_status_code=400,
            response_text='152'
        )

        with self.assertRaises(exceptions.ProxyRackException):
            with RetryClient() as client:
                proxyrack_api.GetCountryIPCount.call(client=client, country='US')

        self.assertEqual(mocked_route_ip_count.call_count, 1)



from py_aws_core.spoofing.exceptions import SpoofingException


class ProxyRackException(SpoofingException):
    ERROR_MESSAGE = 'A ProxyRack Error has occurred'


class ProxyNotAuthenticated(ProxyRackException):
    ERROR_MESSAGE = 'ProxyRack Proxy Not Authenticated'


class GeoLocationNotFound(ProxyRackException):
    ERROR_MESSAGE = 'ProxyRack Geolocation Not Found'


class ProxyUnreachable(ProxyRackException):
    ERROR_MESSAGE = 'ProxyRack Proxy Unreachable'


class ProxyNotFound(ProxyRackException):
    ERROR_MESSAGE = 'ProxyRack Proxy Not Found'


class ProxyNotOnline(ProxyRackException):
    ERROR_MESSAGE = 'ProxyRack Proxy Not Online'



class SpoofingException(Exception):
    ERROR_MESSAGE = 'A generic spoofing error has occurred'

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def __str__(self):
        return self.ERROR_MESSAGE



import typing
from dataclasses import dataclass

from . import exceptions, logs

logger = logs.get_logger()


class APIGatewayRouter:
    """
    Small router for parsing API Gateway Events and calling the matching lambda
    Based on ideas from Tiny-Router
    https://kevinquinn.fun/blog/tiny-python-router-for-aws-lambda/
    """
    VALID_METHODS = ['GET', 'POST', 'PUT', 'DELETE']

    @dataclass
    class PathFuncs:
        fn: typing.Callable
        kwargs: typing.Dict

    def __init__(self):
        self._route_map = dict()

    @property
    def routes(self) -> dict:
        return self._route_map

    def route(self, path: str, http_method: str, **kwargs):
        """
        Decorator to add a route to a function
        :param path:
        :param http_method:
        :param kwargs:
        :return:
        """
        def decorator(fn):
            self.add_route(fn=fn, path=path, http_method=http_method, **kwargs)
            return fn

        return decorator

    def add_route(self, fn: typing.Callable, path: str, http_method: str, **kwargs):
        if http_method not in self.VALID_METHODS:
            raise exceptions.RouteMethodNotAllowed(http_method=http_method, valid_methods=self.VALID_METHODS)
        if http_method in self._route_map and path in self._route_map[http_method]:
            raise exceptions.RouteAlreadyExists(method=http_method, path=path)
        if http_method not in self._route_map:
            self._route_map[http_method] = dict()
        self._route_map[http_method][path] = self.PathFuncs(fn=fn, kwargs=kwargs)
        logger.info(f'Added route to router', http_method=http_method, path=path)

    def handle_event(self, aws_event, aws_context, **kwargs):
        path = aws_event['path']
        http_method = aws_event['httpMethod']
        logger.info(f'Routing event', path=path, http_method=http_method, aws_event=aws_event)
        try:
            path_funcs = self._route_map[http_method][path]
            return path_funcs.fn(aws_event, aws_context, **path_funcs.kwargs, **kwargs)
        except KeyError:
            raise exceptions.RouteNotFound(http_method=http_method, path=path)



import base64
import binascii
import pickle
import ssl
from http.cookiejar import CookieJar

from httpx import Client, HTTPStatusError, TimeoutException, NetworkError, ProxyError

from py_aws_core import decorators, exceptions, logs, utils
from py_aws_core.session_interface import ISession
from py_aws_core.db_service import DBService

logger = logs.get_logger()
# Using same ssl context for all clients to save on loading SSL bundles
# See https://github.com/python/cpython/issues/95031#issuecomment-1749489998
# Also results in _tests running about 9 times faster
SSL_CONTEXT = ssl.create_default_context()


class RetryClient(Client):
    """
    Http/2 Client that retries for given exceptions and http status codes
    """
    RETRY_EXCEPTIONS = (
        HTTPStatusError,
        TimeoutException,
        NetworkError,
        ProxyError
    )

    RETRY_STATUS_CODES = (
        408,
        425,
        429,
        500,
        502,
        503,
        504,
    )

    def __init__(
        self,
        session_id: str = None,
        follow_redirects: bool = True,
        verify: bool = None,
        timeout=10.0,
        *args,
        **kwargs
    ):
        super().__init__(
            follow_redirects=follow_redirects,
            default_encoding="utf-8",
            verify=verify or SSL_CONTEXT,
            timeout=timeout,
            *args,
            **kwargs
        )
        self._session_id = session_id or utils.get_uuid_hex()

    @decorators.retry(retry_exceptions=RETRY_EXCEPTIONS)
    @decorators.http_status_check(reraise_status_codes=RETRY_STATUS_CODES)
    def send(self, *args, **kwargs):
        return super().send(*args, **kwargs)

    @property
    def session_id(self) -> str:
        return self._session_id

    @property
    def b64_encoded_cookies(self) -> bytes:
        return base64.encodebytes(pickle.dumps([c for c in self.cookies.jar]))

    def b64_decode_and_set_cookies(self, b64_cookies: bytes):
        if not b64_cookies:
            logger.info(f'No Cookies To Restore', session_id=self.session_id, b64_cookies=b64_cookies)
            self.cookies.jar = CookieJar()
            return
        try:
            cookie_jar = CookieJar()
            decoded_bytes = base64.decodebytes(b64_cookies)
            for c in pickle.loads(decoded_bytes):
                logger.info(f'Setting CookieJar Cookie: {getattr(c, 'name')}', session_id=self.session_id)
                cookie_jar.set_cookie(c)
            self.cookies.jar = cookie_jar
            logger.info(f'Rehydrated {len(self.cookies.jar)} cookies', session_id=self.session_id)
        except (pickle.PickleError, binascii.Error) as e:
            raise exceptions.CookieDecodingError(info=f'Cookie Error: {str(e)}', session_id=self.session_id)


class SessionPersistClient(RetryClient, ISession):
    def __init__(self, session_service: DBService, *args, **kwargs):
        self._session_service = session_service
        super().__init__(*args, **kwargs)

    def __enter__(self):
        super().__enter__()
        self.read_session()
        return self

    def __exit__(self, *args, **kwargs):
        self.write_session(session_id=self.session_id)
        super().__exit__(*args, **kwargs)

    def read_session(self):
        session_id = self.session_id
        logger.info(f'Attempting to read session...', session_id=self.session_id)
        session = self._session_service.get_or_create_session(session_id=session_id)
        logger.info(f'Successfully read session.', session_id=self.session_id)
        self.b64_decode_and_set_cookies(b64_cookies=session.b64_cookies_bytes)

    def write_session(self, session_id):
        self._session_service.update_session_cookies(session_id=session_id, b64_cookies=self.b64_encoded_cookies)
        logger.info(f'Wrote session cookies to database', session_id=self.session_id)



from abc import ABC

import boto3
from botocore.client import BaseClient
from botocore.config import Config

from .secrets_interface import IDynamoDBSecrets


class ABCBotoSession(ABC):
    CLIENT_CONNECT_TIMEOUT = 4.9
    CLIENT_READ_TIMEOUT = 4.9

    _boto3_session = boto3.Session()

    @classmethod
    def _get_config(cls, **kwargs):
        return Config(
            connect_timeout=cls.CLIENT_CONNECT_TIMEOUT,
            read_timeout=cls.CLIENT_READ_TIMEOUT,
            retries=dict(
                total_max_attempts=2,
            ),
            **kwargs
        )


class ABCBotoClient(ABCBotoSession):
    def __init__(self, service_name: str, verify: bool = True, **kwargs):
        self._boto_client = self._get_new_client(
            service_name=service_name,
            verify=verify,
            **kwargs
        )

    def _get_new_client(self, service_name: str, verify: bool, **kwargs) -> BaseClient:
        return self._boto3_session.client(
            config=self._get_config(),
            service_name=service_name,
            **kwargs
        )

    @property
    def boto_client(self) -> BaseClient:
        return self._boto_client


class ABCBotoResource(ABCBotoSession):
    def __init__(self, service_name: str, **kwargs):
        self._boto_resource = self._get_new_resource(
            service_name=service_name,
            **kwargs
        )

    def _get_new_resource(self, service_name: str, **kwargs) -> BaseClient:
        return self._boto3_session.resource(
            config=self._get_config(),
            service_name=service_name,
            **kwargs
        )

    @property
    def boto_resource(self) -> BaseClient:
        return self._boto_resource


class CognitoClient(ABCBotoClient):
    def __init__(self):
        super().__init__(service_name='cognito-idp')


class SecretManagerClient(ABCBotoClient):
    def __init__(self):
        super().__init__(service_name='secretsmanager')


class SSMClient(ABCBotoClient):
    def __init__(self):
        super().__init__(service_name='ssm')


class DynamoTable(ABCBotoResource):
    def __init__(self, ddb_secrets: IDynamoDBSecrets):
        super().__init__(service_name='dynamodb')
        self._ddb_secrets = ddb_secrets

    @property
    def table(self):
        return self.boto_resource.Table(self.table_name)

    @property
    def table_name(self):
        return self._ddb_secrets.get_table_name()



import random
from abc import ABC, abstractmethod

from py_aws_core import logs
from . import const

logger = logs.get_logger()


class IProxy(ABC):
    def __init__(self, proxy_username: str, proxy_password: str):
        self._proxy_username = proxy_username
        self._proxy_password = proxy_password

    @abstractmethod
    def get_proxy_url(self, **kwargs) -> str:
        raise NotImplemented

    @staticmethod
    def get_weighted_country():
        countries, weights = zip(const.PROXY_COUNTRY_WEIGHTS)
        return random.choices(population=countries, weights=weights, k=1)[0]

    @property
    def proxy_username(self):
        return self._proxy_username

    @property
    def proxy_password(self):
        return self._proxy_password



from py_aws_core import const, db_api, dynamodb_entities, logs, utils
from py_aws_core.db_interface import IDatabase
from py_aws_core.dynamodb_api import DynamoDBAPI

logger = logs.get_logger()


class DBService(IDatabase):
    def __init__(self, table):
        self._table = table

    def get_or_create_session(self, session_id: str) -> dynamodb_entities.Session:
        return db_api.GetOrCreateSession.call(
            table=self._table,
            session_id=session_id,
            created_at_datetime=utils.to_iso_8601(),
            expires_at=DynamoDBAPI.calc_expire_at_timestamp(expire_in_seconds=const.DB_DEFAULT_EXPIRES_IN_SECONDS)
        ).session

    def get_session_item(self, session_id: str) -> dynamodb_entities.Session:
        return db_api.GetSessionItem.call(
            table=self._table,
            session_id=session_id
        ).session

    def put_session_item(self, session_id: str, b64_cookies: bytes):
        return db_api.PutSession.call(
            table=self._table,
            session_id=session_id,
            b64_cookies=b64_cookies
        )

    def update_session_cookies(self, session_id: str, b64_cookies: bytes) -> dynamodb_entities.Session:
        return db_api.UpdateSessionCookies.call(
            table=self._table,
            session_id=session_id,
            b64_cookies=b64_cookies,
            now_datetime=utils.to_iso_8601()
        ).session



import json

from botocore.client import BaseClient
from botocore.exceptions import ClientError

from . import exceptions, logs, utils
from .secrets_interface import ISecrets

logger = logs.get_logger()


class SecretsManager(ISecrets):
    class Response:
        def __init__(self, data):
            self.ARN = data['ARN']
            self.Name = data['Name']
            self.VersionId = data['VersionId']
            self.SecretBinary = data['SecretBinary']
            self.SecretString = data['SecretString']
            self.VersionStages = data['VersionStages']

        @property
        def secret_json(self):
            return json.loads(self.SecretString)

    """
    First checks environment variables for secrets.
    If secret not found, will attempt to pull from secrets manager
    """
    AWS_SECRET_NAME = 'AWS_SECRET_NAME'

    def __init__(self, boto_client: BaseClient):
        self._boto_client = boto_client
        self._secrets_map = dict()

    def get_secret(self, secret_name: str):
        if secret_value := utils.get_environment_variable(secret_name):
            logger.debug(f'Secret "{secret_name}" found in environment variables')
            return secret_value
        if val := self._secrets_map.get(secret_name):
            logger.debug(f'Secret "{secret_name}" found in cached secrets')
            return val
        try:
            r_secrets = self.Response(self._boto_client.get_secret_value(SecretId=self.aws_secret_name))
            self._secrets_map = r_secrets.secret_json
            return self._secrets_map[secret_name]
        except ClientError as e:
            logger.exception(f'Error while trying to find secret "{secret_name}"')
            # For a list of exceptions thrown, see
            # https://docs.aws.amazon.com/secretsmanager/latest/apireference/API_GetSecretValue.html
            raise exceptions.SecretsManagerException(e)

    @property
    def aws_secret_name(self) -> str:
        if aws_secret_id := utils.get_environment_variable(self.AWS_SECRET_NAME):
            return aws_secret_id
        raise exceptions.SecretsManagerException(f'Missing environment variable "{self.AWS_SECRET_NAME}"')



import typing

from py_aws_core import logs
from py_aws_core.spoofing.proxy_interface import IProxy
from py_aws_core.spoofing.proxyrack import const, utils

logger = logs.get_logger()


class ProxyRackService(IProxy):
    def get_proxy_url(
        self,
        netloc: str,
        cities: typing.List[str] = None,
        country: str = None,
        isps: typing.List[str] = None,
        proxy_ip: str = None,
        proxy_os: const.ProxyOs = None,
        session_id: str = None,
        **kwargs
    ) -> str:
        config = utils.ProxyBuilder.Config(
            cities=cities,
            country=country,
            isps=isps,
            proxy_ip=proxy_ip,
            proxy_os=proxy_os,
            session_id=session_id,
            refresh_minutes=60
        )
        return utils.ProxyBuilder(
            username=self.proxy_username,
            password=self.proxy_password,
            netloc=netloc,
            config=config
        ).http_url



import typing
from abc import ABC
from dataclasses import dataclass
from enum import Enum

from botocore.client import BaseClient

from py_aws_core import decorators, exceptions, logs, mixins

logger = logs.get_logger()


class AuthChallenge(Enum):
    SMS_MFA = 'SMS_MFA'
    EMAIL_OTP = 'EMAIL_OTP'
    SOFTWARE_TOKEN_MFA = 'SOFTWARE_TOKEN_MFA'
    SELECT_MFA_TYPE = 'SELECT_MFA_TYPE'
    MFA_SETUP = 'MFA_SETUP'
    PASSWORD_VERIFIER = 'PASSWORD_VERIFIER'
    CUSTOM_CHALLENGE = 'CUSTOM_CHALLENGE'
    DEVICE_SRP_AUTH = 'DEVICE_SRP_AUTH'
    DEVICE_PASSWORD_VERIFIER = 'DEVICE_PASSWORD_VERIFIER'
    ADMIN_NO_SRP_AUTH = 'ADMIN_NO_SRP_AUTH'
    NEW_PASSWORD_REQUIRED = 'NEW_PASSWORD_REQUIRED'


class AdminCreateUser:
    class Response:
        class User:
            class MFAOptions:
                def __init__(self, data: dict):
                    self.DeliveryMedium = data.get('DeliveryMedium')
                    self.AttributeName = data.get('AttributeName')

            class Attribute:
                def __init__(self, data: dict):
                    self.Name = data.get('Name')
                    self.Value = data.get('Value')

            def __init__(self, data: dict):
                self.Username = data.get('Username')
                self.Attributes = [self.Attribute(a) for a in data.get('Attributes')]
                self.UserCreateDate = data.get('UserCreateDate')
                self.UserLastModifiedDate = data.get('UserLastModifiedDate')
                self.Enabled = data.get('Enabled')
                self.UserStatus = data.get('UserStatus')
                self.MFAOptions = [self.MFAOptions(mfa) for mfa in data.get('MFAOptions')]

        def __init__(self, data: dict):
            self.User = self.User(data.get('User', dict()))

    @classmethod
    @decorators.boto3_handler(raise_as=exceptions.CognitoException)
    def call(
        cls,
        cognito_client: BaseClient,
        cognito_pool_id: str,
        username: str,
        user_attributes: typing.List[typing.Dict],
        desired_delivery_mediums: typing.List[str],
    ):
        response = cognito_client.admin_create_user(
            DesiredDeliveryMediums=desired_delivery_mediums,
            Username=username,
            UserAttributes=user_attributes,
            UserPoolId=cognito_pool_id
        )
        logger.info('Cognito AdminCreateUser called', response=response)
        return cls.Response(response)


class ABCInitiateAuth(ABC):
    class Response:
        class AuthenticationResult:
            class NewDeviceMetadata:
                def __init__(self, data: dict):
                    self.DeviceKey = data.get('DeviceKey', dict())
                    self.DeviceGroupKey = data.get('DeviceGroupKey', dict())

            def __init__(self, data: dict):
                self._data = data
                self.AccessToken = data.get('AccessToken')
                self.ExpiresIn = data.get('ExpiresIn')
                self.TokenType = data.get('TokenType')
                self.RefreshToken = data.get('RefreshToken')
                self.IdToken = data.get('IdToken')
                self.NewDeviceMetadata = self.NewDeviceMetadata(data.get('NewDeviceMetadata', dict()))

        def __init__(self, data: dict):
            self.ChallengeName = data.get('ChallengeName')
            self.Session = data.get('Session')
            self.ChallengeParameters = data.get('ChallengeParameters')
            self.AuthenticationResult = self.AuthenticationResult(data.get('AuthenticationResult', dict()))


class UserPasswordAuth(ABCInitiateAuth):
    @classmethod
    @decorators.boto3_handler(raise_as=exceptions.CognitoException)
    def call(
        cls,
        cognito_client: BaseClient,
        cognito_pool_client_id: str,
        username: str,
        password: str,

    ):
        response = cognito_client.initiate_auth(
            AuthFlow='USER_PASSWORD_AUTH',
            AuthParameters={
                'USERNAME': username,
                'PASSWORD': password,
            },
            ClientId=cognito_pool_client_id,
        )
        logger.info('Cognito UserPasswordAuth called', response=response)
        return cls.Response(response)


class RefreshTokenAuth(ABCInitiateAuth):
    @classmethod
    @decorators.boto3_handler(raise_as=exceptions.CognitoException)
    def call(
        cls,
        cognito_client: BaseClient,
        cognito_pool_client_id: str,
        refresh_token: str,
    ):
        response = cognito_client.initiate_auth(
            AuthFlow='REFRESH_TOKEN',
            AuthParameters={
                'REFRESH_TOKEN': refresh_token,
            },
            ClientId=cognito_pool_client_id,
        )
        logger.info('Cognito RefreshTokenAuth called', response=response)
        return cls.Response(response)


@dataclass
class ABCChallengeResponse(ABC, mixins.AsDictMixin):
    pass


@dataclass
class NewPasswordChallengeResponse(ABCChallengeResponse):
    NEW_PASSWORD: str
    USERNAME: str
    user_attributes: dict

    def __post_init__(self):
        self.set_attrs_from_user_attributes()

    def set_attrs_from_user_attributes(self):
        for k, v in self.user_attributes.items():
            self.__dict__[f'userAttributes.{k}'] = v
        self.__dict__.pop('user_attributes')


class RespondToAuthChallenge(ABCInitiateAuth):
    @classmethod
    @decorators.boto3_handler(raise_as=exceptions.CognitoException)
    def call(
        cls,
        cognito_client: BaseClient,
        cognito_pool_client_id: str,
        challenge_name: AuthChallenge,
        challenge_responses: ABCChallengeResponse,
        session: str = '',
    ):
        response = cognito_client.respond_to_auth_challenge(
            ChallengeName=challenge_name.value,
            ChallengeResponses=challenge_responses.as_dict(),
            ClientId=cognito_pool_client_id,
            Session=session,
        )
        logger.info('Cognito RespondToAuthChallenge called', response=response)
        return cls.Response(response)



import importlib
import json
import os
import random
import time
import typing
import uuid
from datetime import datetime, timezone, timedelta, UTC
from importlib.resources import files
from pathlib import Path


def build_lambda_response(
    status_code: int,
    body: typing.Dict | str = None,
    content_type: str = 'application/json',
    multi_value_headers: typing.Dict[str, typing.List[str]] = None,
    exc: Exception = None
):
    """
      Note - CORS response headers will NOT be present in response if using proxy type integration.
      CORS response headers must be set in lambda response
      See below for further details:
      https://stackoverflow.com/a/54089431
      https://docs.aws.amazon.com/apigateway/latest/developerguide/how-to-cors.html#apigateway-enable-cors-proxy

    :param status_code:
    :param body:
    :param content_type:
    :param multi_value_headers:
    :param exc:
    :return:
    """
    if not body:
        body = dict()
    if exc:
        body['error'] = {
            'type': type(exc).__name__,
            'message': str(exc),
        }
    if isinstance(body, dict):
        body = json.dumps(body)
    response_headers = {
        'Content-Type': [content_type],
        'Access-Control-Allow-Credentials': [True],
        'Access-Control-Allow-Headers': ['Content-Type,X-Amz-Date,Authorization,X-Api-Key,X-Amz-Security-Token'],
        'Access-Control-Allow-Origin': ['*'],
        'Access-Control-Allow-Methods': ['DELETE,GET,POST,PUT']
    }
    if multi_value_headers:
        response_headers |= multi_value_headers
    return {
        'isBase64Encoded': False,
        'statusCode': status_code,
        'body': body,  # body key value must be a json string
        'multiValueHeaders': response_headers
    }


def to_iso_8601(dt: datetime = None, tz=timezone.utc) -> str:
    """
    Example output 2020-07-10 15:00:00.000
    :param tz:
    :param dt:
    :return:
    """
    if not dt:
        dt = get_now_datetime(tz=tz).replace(microsecond=0)
    return dt.isoformat()


def add_seconds_to_current_unix_timestamp(seconds: int, tz=timezone.utc) -> int:
    dt = get_now_datetime(tz=tz) + timedelta(seconds=seconds)
    return int(dt.timestamp())


def add_seconds_to_now_datetime(seconds: int, tz=timezone.utc) -> datetime:
    return get_now_datetime(tz=tz) + timedelta(seconds=seconds)


def get_now_datetime(tz=timezone.utc) -> datetime:
    return datetime.now(tz=tz)


def generate_jitter(midpoint: float, floor: float, std_deviation: float) -> float:
    return max(floor, midpoint + random.uniform(-std_deviation, +std_deviation))


def rand_int(num_a: int, num_b: int) -> int:
    return random.randint(num_a, num_b)


def sleep(seconds: float) -> None:
    return time.sleep(seconds)


def get_uuid_hex() -> str:
    return uuid.uuid4().hex


def get_environment_variable(secret_name: str, default=None) -> str:
    return os.environ.get(secret_name, default=default)


def decode_unicode(s: str) -> str:
    """
    Per https://stackoverflow.com/a/33134946
    This is how to fix unicode decoding errors
    :param s: string to decode
    :return:
    """
    try:
        return s.encode('iso-8859-1').decode('utf-8')
    except (UnicodeEncodeError, UnicodeDecodeError):
        return s


def unix_timestamp_to_iso8601(unix_ts: int):
    return datetime.fromtimestamp(unix_ts, tz=UTC).isoformat()


def remove_whitespace(s: str) -> str:
    return ''.join([line.strip() for line in s.split('\\r\\n')])


def to_utf8_bytes(s: str) -> bytes:
    return bytes(s, 'utf-8')


def import_all_package_modules(package: str):
    f = files(package)
    modules = [fp for fp in f.iterdir() if fp.is_file and fp.name.endswith('.py')]
    for fp in modules:
        module_name = Path(fp.name).stem
        full_name = f'{package}.{module_name}'
        try:
            importlib.import_module(full_name)
        except ModuleNotFoundError:
            continue



import typing

from py_aws_core.spoofing.proxyrack import const


class ProxyBuilder:

    class Config:
        def __init__(
            self,
            cities: typing.List[str] = None,
            country: str = None,
            isps: typing.List[str] = None,
            proxy_ip: str = None,
            proxy_os: const.ProxyOs = None,
            refresh_minutes: int = 10,
            session_id: str | str = None,
            **kwargs
        ):
            self._cities = cities
            self._country = country
            self._isps = isps
            self._proxy_ip = proxy_ip
            self._proxy_os = proxy_os
            self._session_id = session_id
            self._refresh_minutes = refresh_minutes
            self._kwargs = kwargs

        @property
        def to_string(self) -> str:
            opts = {
                'city': self.to_str(self._cities),
                'country': self._country,
                'isp': self.to_str(self._isps),
                'osName': self._proxy_os.value,
                'proxyIp': self._proxy_ip,
                'session': self._session_id,
                'refreshMinutes': self._refresh_minutes
            } | self._kwargs
            return ';'.join([f'{k}={str(v).replace(" ", "")}' for k, v in opts.items() if v])

        @staticmethod
        def to_str(objs: typing.List[str]) -> str:
            if objs:
                return ','.join(objs)
            return ''

    def __init__(self, username: str, password: str, netloc: str, config: Config):
        self._username = username
        self._password = password
        self._netloc = netloc
        self._config = config

    @property
    def http_url(self) -> str:
        return f'http://{self.params}:{self._password}@{self._netloc}'

    @property
    def params(self):
        return ';'.join([self._username, self._config.to_string])



import json
import time
from http.cookiejar import Cookie
from importlib.abc import Traversable
from importlib.resources import as_file
from importlib.resources import files
from unittest import TestCase

import respx
from httpx import Response, codes

from . import utils
from .secrets_interface import IDynamoDBSecrets


class BaseTestFixture(TestCase):

    TEST_AUTH_TOKEN = 'eyJraWQiOiJzNndrcytDXC84WGxNOVF2OHNYeVhGczNjV1VsOVFwVzZsdE9rMGt5R2dDVT0iLCJhbGciOiJSUzI1NiJ9.eyJzdWIiOiIzZTYyYzNjOC00OTcwLTRmYzctYTk5Ni04NjhkNGMxMzk3ZjYiLCJjdXN0b206cm9sZXMiOiJTVVBFUlVTRVIsTUVNQkVSLFNUQUZGIiwiZW1haWxfdmVyaWZpZWQiOnRydWUsImlzcyI6Imh0dHBzOlwvXC9jb2duaXRvLWlkcC51cy13ZXN0LTIuYW1hem9uYXdzLmNvbVwvdXMtd2VzdC0yX0N4VlFXNU1wVSIsImN1c3RvbTpncm91cCI6Indob2FkZXJlanIiLCJjb2duaXRvOnVzZXJuYW1lIjoiaGVsbG9tb3RvIiwib3JpZ2luX2p0aSI6IjZmMjA5YzgyLWMyNTgtNGIzMC1hMGIxLTBjYWYwMDRkMDRkNiIsImF1ZCI6IjdtdWR1dGdiZGViY2g2YWVoMjF1ZXEyaDFtIiwiZXZlbnRfaWQiOiJhMGNhODUwOS0xMTM0LTRmZWEtYTg5YS1kN2ZjYjRmOTg3ZTciLCJ0b2tlbl91c2UiOiJpZCIsImF1dGhfdGltZSI6MTcxMzY2NDMyMCwiZXhwIjoxNzEzNjY3OTIwLCJpYXQiOjE3MTM2NjQzMjAsImp0aSI6IjgzNGI3ZWYyLTBkN2EtNGM2My1hYjI1LWMxYjU1MTU5M2VhNiIsImVtYWlsIjoiYWRyaWFuQHJ5ZGVhcy5jb20ifQ.HrlCeTFFfY_lKtA8HFFrzWRdyCdWbNdaMRozDH65Uzyy9FaavZq2enU8lQSs_TOyEl9aj5mxdIOmN03bpsaiZeWFLD6Ph_oUHl6MQBlgMGPQjzamAO6homn2IH-Thn5jcFdUVWEHIA4GHrw2M_Rty5sCDBxVkdjqCqU4KchoxL2zwXCGdp4Fr-p22PkG8bAkrLQAtO_QE2HqrseJo0XkOm64GDNVvNbL4O4EJ3NUY3vkRck2L0g0GMYG_1x8GFgwr1MdFOg0BOXvL4Qu0ToEVK86mjL-Ua8MR9t9Z4VraKIU5RSncm3SWuWTW6K2uo1oeXAXNWWhrA01Q3BeTGVvJg'
    TEST_RESOURCE_PATH = 'tests._resources'
    TEST_COGNITO_POOL_CLIENT_ID = '7xxxxtgbdebch6fffffueq2h1m'
    TEST_COGNITO_POOL_ID = 'us-west-2_CCCQW5ZZZ'
    TEST_ACCESS_TOKEN = 'eifuhwseduivfavhwveci'
    TEST_SESSION = 'AYABeFXvmcOOYfDt0FEXFF1YJU0AHQABAAdTZXJ2aWNlABBDb2duaXRvVXNlclBvb2xzAAEAB2F3cy1rbXMAS2Fybjphd3M6a21zOnVzLXdlc3QtMjowMTU3MzY3MjcxOTg6a2V5LzI5OTFhNGE5LTM5YTAtNDQ0Mi04MWU4LWRkYjY4NTllMTg2MQC4AQIBAHhPj7k9zU4nGXUQUvM0Ccwk42DS-fm3vKmH75ktTrktNQGm02KVuFeo-2uJN5UBn74vAAAAfjB8BgkqhkiG9w0BBwagbzBtAgEAMGgGCSqGSIb3DQEHATAeBglghkgBZQMEAS4wEQQMV7Hs08X6HB6WKPsgAgEQgDuSOHzEU5wh_FikpJYjxXGPbYB3BC_fEDb05YWJnrokkGDf12Yrg53fxiUDNW7B523L2axIC7FT9tMShAIAAAAADAAAEAAAAAAAAAAAAAAAAADEdVndAEPLobnWu0H25hNC_____wAAAAEAAAAAAAAAAAAAAAEAAADT8xRI6YxhCSqev-6ZOdk2vuTmWFPCNGyhmbIKRG2yc2OVcWkuNYnX1neWZXYIFy4bmUseiL6_Bfr21wCg8JEnkgoNrc6crlRWSADrc3N7tVuPP_N_CbYPVX2_NurbuigG68m6mxWI9gXiep5W2OuqJWC_RUQlFrYvUDnMv0EiYHxxzBun7L3wG5uKa7UemhNbEU0UFKpLnbvFyT1Vk0UUzIIQVEalkHw7flr2cKthKoAAdHb1moJmGTxVPAG1FCrxqYL9C7HtKwOk2WLRuRw5_298fyLsJwRmR1CwM0L_S4IVor4'

    TEST_BOTO3_RESOURCES_PATH = files(TEST_RESOURCE_PATH).joinpath('boto3')
    TEST_BOTO3_ERROR_RESOURCES_PATH = TEST_BOTO3_RESOURCES_PATH.joinpath('errors')

    TEST_COGNITO_RESOURCES_PATH = files(TEST_RESOURCE_PATH).joinpath('cognito')
    TEST_COGNITO_ERROR_RESOURCES_PATH = TEST_COGNITO_RESOURCES_PATH.joinpath('errors')

    TEST_DB_RESOURCES_PATH = files(TEST_RESOURCE_PATH).joinpath('db')
    TEST_DB_ERROR_RESOURCES_PATH = TEST_DB_RESOURCES_PATH.joinpath('errors')

    TEST_LAMBDA_RESOURCES_PATH = files(TEST_RESOURCE_PATH).joinpath('lambda')
    TEST_LAMBDA_ERROR_RESOURCES_PATH = TEST_DB_RESOURCES_PATH.joinpath('errors')

    TEST_SECRETS_MANAGER_RESOURCES_PATH = files(TEST_RESOURCE_PATH).joinpath('secrets_manager')
    # TEST_SECRETS_MANAGER_RESOURCES_PATH = TEST_DB_RESOURCES_PATH.joinpath('errors')

    TEST_SSM_RESOURCES_PATH = files(TEST_RESOURCE_PATH).joinpath('ssm')

    def setUp(self):
        self.start_time = time.time()
        super().setUp()

    def tearDown(self):
        total = time.time() - self.start_time
        if total > 0.1:
            print(f'{self.id} slow test took {total:.3f} seconds')
        super().tearDown()

    @classmethod
    def create_ok_route(cls, method='GET', headers=None, _json=None, text=None, **kwargs):
        return cls.create_route(
            method=method,
            response_status_code=codes.OK,
            response_headers=headers,
            response_json=_json,
            response_text=text,
            **kwargs
        )

    @classmethod
    def create_bad_request_route(cls, **kwargs):
        return cls.create_route(response_status_code=codes.BAD_REQUEST, **kwargs)

    @classmethod
    def create_route(
        cls,
        response_status_code,
        method='GET',
        response_headers=None,
        response_json=None,
        response_text=None,
        **kwargs
    ):
        return respx.route(
            method=method,
            **kwargs
        ).mock(
            return_value=Response(
                headers=response_headers,
                status_code=response_status_code,
                json=response_json,
                text=response_text,
            )
        )

    @classmethod
    def create_test_cookie(cls, name: str, value: str) -> Cookie:
        return Cookie(
            version=1,
            name=name,
            value=value,
            port=None,
            port_specified=False,
            domain='www.example.com',
            domain_specified=True,
            domain_initial_dot=False,
            path='/',
            path_specified=True,
            secure=True,
            expires=3600,
            discard=False,
            comment=None,
            comment_url=None,
            rest=dict(),
        )

    @classmethod
    def to_utf8_bytes(cls, s: str):
        return utils.to_utf8_bytes(s)

    @classmethod
    def get_resource_json(cls, *descendants: str, path: Traversable,):
        source = path.joinpath(*descendants)
        with as_file(source) as file_text:
            return json.loads(file_text.read_text(encoding='utf-8'))

    @classmethod
    def get_resource_text(cls, *descendants: str, path: Traversable):
        source = path.joinpath(*descendants)
        with as_file(source) as file_text:
            return file_text.read_text(encoding='utf-8')

    class MockDynamoDBSecretsService(IDynamoDBSecrets):
        def get_secret(self, secret_name: str) -> str:
            pass

        def get_table_name(self) -> str:
            return 'TEST_TABLE'



import typing
from abc import ABC, abstractmethod

from py_aws_core import encoders
from py_aws_core.dynamodb_api import DynamoDBAPI


class BaseModel:
    def __init__(self, data):
        self.PK = data.get('PK')
        self.SK = data.get('SK')
        self.Type = data.get('Type')
        self.CreatedAt = data.get('CreatedAt')
        self.CreatedBy = data.get('CreatedBy')
        self.ModifiedAt = data.get('ModifiedAt')
        self.ModifiedBy = data.get('ModifiedBy')
        self.ExpiresAt = data.get('ExpiresAt')

    @property
    def to_json(self):
        return encoders.JsonEncoder().serialize_to_json(self)

    @staticmethod
    def deserialize_data(data: dict) -> typing.Dict:
        return DynamoDBAPI.deserialize_types(data)


class ABCEntity(ABC, BaseModel):
    TYPE = 'ABC'

    @classmethod
    @abstractmethod
    def create_key(cls, *args, **kwargs) -> str:
        pass

    @classmethod
    def type(cls) -> str:
        return cls.TYPE


class Session(ABCEntity):
    TYPE = 'SESSION'

    def __init__(self, data):
        super().__init__(data)
        self.Base64Cookies = data.get('Base64Cookies')
        self.SessionId = data.get('SessionId')

    @classmethod
    def create_key(cls, _id: str) -> str:
        return f'{cls.type()}#{str(_id)}'

    @property
    def b64_cookies_bytes(self):
        if self.Base64Cookies:
            return self.Base64Cookies.value
        return None



import logging
import os
import sys

import structlog
from dotenv import load_dotenv

dotenv = load_dotenv()  # take environment variables from .env.
level = os.environ.get("LOG_LEVEL", "INFO").upper()
LOG_LEVEL = getattr(logging, level)

exception_transformer = structlog.processors.ExceptionDictTransformer(locals_max_length=10000, locals_max_string=10000)

shared_processors = [
    structlog.processors.EventRenamer('message'),
    structlog.processors.CallsiteParameterAdder(),
    structlog.processors.TimeStamper(fmt='iso'),
    structlog.processors.add_log_level,
    structlog.processors.ExceptionPrettyPrinter(exception_formatter=exception_transformer),
]
val = sys.stderr.isatty()
if 'unittest' in sys.modules or sys.stderr.isatty():
    # Pretty printing when we run tests or a terminal session.
    # Automatically prints pretty tracebacks when "rich" is installed
    processors = shared_processors + [
        structlog.dev.ConsoleRenderer(
            colors=True,
            event_key='message'
        ),
    ]
else:
    # Print JSON when we run, e.g., in a Docker container.
    # Also print structured tracebacks.
    processors = shared_processors + [
        structlog.processors.dict_tracebacks,
        structlog.processors.JSONRenderer(),
    ]
structlog.configure(
    processors=processors,
    wrapper_class=structlog.make_filtering_bound_logger(LOG_LEVEL)
)

__logger = structlog.get_logger(__name__)


def get_logger():
    return __logger



import typing


class ErrorResponse:
    class Error:
        def __init__(self, data):
            self.Message = data['Message']
            self.Code = data['Code']

    class CancellationReason:
        def __init__(self, data):
            self.Code = data['Code']
            self.Message = data.get('Message')

    def __init__(self, data):
        self.Error = self.Error(data.get('Error', dict()))
        self.ResponseMetadata = ResponseMetadata(data.get('ResponseMetadata', dict()))
        self.Message = data.get('Message')
        self.CancellationReasons = [self.CancellationReason(r) for r in data.get('CancellationReasons', list())]

    def raise_for_cancellation_reasons(self, error_maps: list[dict[str, typing.Any]]):
        for reason, error_map in zip(self.CancellationReasons, error_maps):
            if exc := error_map.get(reason.Code):
                raise exc


class ItemResponse:
    def __init__(self, data):
        self.Item = data.get('Item')
        self.ResponseMetadata = ResponseMetadata(data.get('ResponseMetadata', dict()))


class QueryResponse:
    def __init__(self, data):
        self._items = data.get('Items') or list()
        self.count = data.get('Count')
        self.scanned_count = data.get('ScannedCount')
        self.response_metadata = ResponseMetadata(data['ResponseMetadata'])

    def get_by_type(self, _type: str) -> list:
        if self._items:
            return [i for i in self._items if i['Type']['S'] == _type]
        return list()


class UpdateItemResponse:
    def __init__(self, data: dict):
        self.attributes = data['Attributes']


class TransactionResponse:
    def __init__(self, data):
        self._data = data
        self.Responses = data.get('Responses')

    @property
    def data(self):
        return self._data


class ResponseMetadata:
    class HTTPHeaders:
        def __init__(self, data):
            self.server = data.get('server')
            self.date = data.get('date')

    def __init__(self, data):
        self.RequestId = data.get('RequestId')
        self.HTTPStatusCode = data.get('HTTPStatusCode')
        self.HTTPHeaders = self.HTTPHeaders(data.get('HTTPHeaders', dict()))
        self.RetryAttempts = data.get('RetryAttempts')



import typing
from functools import wraps
from typing import Any, Type

from botocore.exceptions import ClientError
from httpx import codes, HTTPStatusError

from py_aws_core import exceptions, logs, utils
from py_aws_core.boto_responses import ErrorResponse

logger = logs.get_logger()


def boto3_handler(raise_as: Type[exceptions.CoreException]):
    def deco_func(func):
        @wraps(func)
        def wrapper_func(*args, **kwargs):
            try:
                response = func(*args, **kwargs)
                logger.debug(f'boto response', response=response, wrapped_func_name=f'{func!r}')
                return response
            except ClientError as e:
                error = e.response['Error']
                message = error['Message']
                logger.error(f'Boto client error', code=error['Code'], message=message, response=e.response)
                raise raise_as(message=message)
            except Exception:  # Raise all other exceptions as is
                raise
        return wrapper_func  # true decorator
    return deco_func


def dynamodb_handler(client_err_map: dict[str, Any], cancellation_err_maps: list[dict[str, Any]]):
    def deco_func(func):
        @wraps(func)
        def wrapper_func(*args, **kwargs):
            try:
                response = func(*args, **kwargs)
                logger.debug(f'response from dynamodb', response=response, wrapped_func_name=f'{func!r}')
                return response
            except ClientError as e:
                logger.error(f'dynamodb ClientError detected', e=e, response=e.response, wrapped_func_name=f'{func!r}')
                e_response = ErrorResponse(e.response)
                if e_response.CancellationReasons:
                    e_response.raise_for_cancellation_reasons(error_maps=cancellation_err_maps)
                if exc := client_err_map.get(e_response.Error.Code):
                    raise exc(e)
                raise
        return wrapper_func  # true decorator
    return deco_func


def lambda_response_handler(raise_as: Type[exceptions.CoreException]):
    """
    Handler for any exceptions raised by wrapped function
    Any uncaught exceptions are wrapped and re-raised as the "raise_as" parameter
    Standard lambda response is returned
    :param raise_as:
    :return:
    """
    def deco_func(func):
        @wraps(func)
        def wrapper_func(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except raise_as as e:
                exc = e
                logger.exception('Pass-Through exception detected')
            except Exception as e:
                exc = raise_as(exc=e)
                logger.exception('Fall-Through exception detected')  # Note:  # the logging.exception method just inside the except part
            return utils.build_lambda_response(
                status_code=exc.HTTP_STATUS_CODE,
                exc=exc
            )
        return wrapper_func  # true decorator
    return deco_func


def retry(
    retry_exceptions: typing.Tuple,
    tries: int = 4,
    delay: float = 1.5,
    backoff: float = 2.0,
    jitter: float = 0.1
):
    """
    Retry calling the decorated function using an exponential backoff.

    Args:
        :param retry_exceptions: Exceptions to retry on before raise
        :param tries: Number of times to try (not retry) before giving up.
        :param delay: Initial delay between retries in seconds.
        :param backoff: Backoff multiplier (e.g. value of 2.0 will double the delay each retry).
        :param jitter: adds a standard deviation to delay
    """
    def deco_func(func):

        @wraps(func)
        def wrapper_func(*args, **kwargs):
            m_tries = 1
            m_delay = delay
            while m_tries < tries:
                try:
                    return func(*args, **kwargs)
                except retry_exceptions as e:
                    j_delay = utils.generate_jitter(midpoint=m_delay, floor=0, std_deviation=jitter)
                    logger.info(
                        f'Retrying in {j_delay:.3f} seconds...',
                        num_tries=m_tries,
                        max_tries=tries,
                        exception=str(e),
                        exception_type=type(e),
                        wrapped_func_name=f'{func!r}'
                    )
                    utils.sleep(j_delay)
                    m_tries += 1
                    m_delay *= backoff

            logger.warning(f'Max tries reached', num_tries=m_tries, max_tries=tries, wrapped_func_name=f'{func!r}')
            return func(*args, **kwargs)

        return wrapper_func  # true decorator

    return deco_func


def http_status_check(reraise_status_codes: typing.Tuple[int, ...] = tuple()):
    def deco_func(func):
        @wraps(func)
        def wrapper_func(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except HTTPStatusError as e:
                status_code = e.response.status_code
                if status_code in reraise_status_codes:     # Retryable status codes
                    raise
                if status_code == codes.UNAUTHORIZED:
                    raise exceptions.NotAuthorizedException(**kwargs, **e.__dict__)
                raise exceptions.APIException(**kwargs, **e.__dict__)

        return wrapper_func

    return deco_func


def wrap_exceptions(raise_as: Type[exceptions.CoreException]):

    def deco_func(func):

        @wraps(func)
        def wrapper_func(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except raise_as:
                raise
            except Exception as e:
                raise raise_as(**kwargs, **e.__dict__)

        return wrapper_func  # true decorator

    return deco_func
